/* 
*	Module: Irregular Network Support Main Module: 'dd.c'
*
*	Contains the following:
*
*	- Routines for Building and Maintaining a Resource Dependency Graph		
*	- Routines for Detecting Cycles in the Resource Dependency Graph
*	- Routines for Detecting Deadlocks Involving Groups of Cycles 
*/


#include 	<stdio.h>
#include	<malloc.h>
#include	"dat2.h"
#include	"dd.h"		
#include	"irnet.h"

double drand48();

/*
*	
*	G L O B A L   V A R I A B L E S 
*
*/


/*
*	Global variable declarations for the adjacency list 
*	(which represents the irregular network topology)
*/

struct	MAINNETNODE	MainNetNodeList[MAXNODES];	/* List of Circuits */

struct TABLEENTRY	*LookUp;


/*
*	Global variable declarations for queue management
*/

int	HeadOfQueue;
int	TailOfQueue;
int	*Queue;



/*
*	Global variable declarations for Up/Down Lookup Table
*/

struct TABLEENTRY	*UDLookUp;


/*
*	Global variable declarations for Up/Down queue management
*/

int	HeadOfUDQueue;
int	TailOfUDQueue;
struct 	UDQUEUEENTRY	*UDQueue;




/* 
*	External variables (from FlexSim) 
* 
*/

extern int IRNUMOFNODES;
extern int IRNUMOFLINKS;
extern struct plinkEnd *plinkTable;




/* 
*
*	  I N I T I A L I Z A T I O N   R O U T I N E S 
*
*/

/*
*	InitMainNetNodeList
*
*	Initializes all values and pointers in the Main Network Node List.
*	(used to initialize this structure before it is used)
*	
*/

void	InitMainNetNodeList()
{
	int	i;

	for (i = 0; i < MAXNODES; i++)
	{
		MainNetNodeList[i].NodeID = i;		/* Not needed ? */	
		MainNetNodeList[i].Degree = 0;	
		MainNetNodeList[i].Color = WHITE;	
		MainNetNodeList[i].Level = UNDEFINED_LEVEL;	
		MainNetNodeList[i].HotSpot = FALSE;	
		MainNetNodeList[i].FirstNetNode = NULL_;
	}
}


/*
*	FreeMainNetNodeListEntries
*
*	Frees the memory allocated to items in the list.
*	(used for "cleanup" purposes.)
*	
*/

void	FreeMainNetNodeListEntries()
{
	int	i;
	NETNODEPTR	CurrentNetNode, NextNetNode;

	for (i = 0; i < MAXNODES; i++)
	{
		CurrentNetNode = MainNetNodeList[i].FirstNetNode;

		while (CurrentNetNode != NULL_)
		{
			NextNetNode = CurrentNetNode->NextNetNode;

			free(CurrentNetNode);

			CurrentNetNode = NextNetNode;
		}
	}
}


/*
*	InitColor
*
*	Initializes the Color parameter of adjacency list.
*	Should be used before each invocation of Breadth First Search.
*/

void	InitColor()
{
	int	i;

	for (i = 0; i < IRNUMOFNODES; i++)
	{
		MainNetNodeList[i].Color = WHITE;	
	}
}


/* 
*
*	  P R I N T   ( D E B U G G I N G )   R O U T I N E S 
*
*/

/*
*	PrintIRNetAdjacencyList
*
*	Prints out the contents of the Adjacency List.
*	Intended to be used for testing/debugging.
*
*/

void	PrintIRNetAdjacencyList()
{
	int		i;
	NETNODEPTR	CurrentNetNode;


	printf("*** IRREGULAR NETWORK ADJACENCY LIST:\n");

	for (i = 0; i < IRNUMOFNODES; i++)
	{
		/* print out the info. regarding circuits */
		printf("[ N: %d | Deg:%d | Col:%d | Lev:%d | HS:%d ] ->", 
			MainNetNodeList[i].NodeID, 
			MainNetNodeList[i].Degree,
			MainNetNodeList[i].Color,
			MainNetNodeList[i].Level,
			MainNetNodeList[i].HotSpot
		);

		CurrentNetNode = MainNetNodeList[i].FirstNetNode;

		/* print out the list of VC nodes */
		while (CurrentNetNode != NULL_) 
		{

			printf("[ N: %d | P: %d | C: %d ] ->", 
				CurrentNetNode->NodeID, 
				CurrentNetNode->PhysLink, 
				CurrentNetNode->Connection);

			CurrentNetNode = (CurrentNetNode->NextNetNode);
		}

		printf(" NULL\n");

	}
	printf("\n\n");
}


/*
*	PrintRoutingTable
*
*	Prints out the contents of the Routing Table.
*
*	'TableType' should be one of the following values:
*
*	STATIC: causes printing out table in single entry per 
*		current node/dest. node pair (same format
*		as the input file when manually specifying static 
*		routing table).
*
*	ADAPTIVE: causes printing out table in multiple entries per 
*		current node/dest (enclosed in brackets, separated by commas).
*		Should be same as the input file when manually specifying 
*		adaptive routing table).
*
*	UP_DOWN: causes printing out table in multiple entries per 
*		current node/dest (enclosed in brackets, separated by commas).
*		Note this is similar, but a separate structure from the adaptive
*		routing table.  
*/

void	PrintRoutingTable(TableType)
int	TableType;
{
	int	i, j;
	int	Index;
	CONNLISTITEMPTR	CurrentConnection;

	printf("*** ROUTING TABLE:\n");

	for( i=0 ; i < IRNUMOFNODES; i++)
	{
		for( j=0 ; j < IRNUMOFNODES; j++)
		{
			Index = (i * IRNUMOFNODES) + j;

			/* Print single entry of the static table */
			if (TableType == STATIC)
			{
				if (i == j)
					printf("X ");
				else
					printf("%d ", LookUp[Index].StaticConnectionID);
			}

			/* Print multiple entries of the adaptive table */
			else if (TableType == ADAPTIVE)
			{
				CurrentConnection = LookUp[Index].FirstConnection;

				if (CurrentConnection == NULL_)
				{
					printf("[X]");
				}
				else
				{
					printf("[");

					do
					{
						if (CurrentConnection->NextConnection == LookUp[Index].FirstConnection)
							printf("%d]", CurrentConnection->ConnectionID);
						else
							printf("%d,", CurrentConnection->ConnectionID);

						CurrentConnection = CurrentConnection->NextConnection;
					}
					while (CurrentConnection != LookUp[Index].FirstConnection);
				}
			}

			/* Print multiple entries of the Up/Down table */
			else /* if (TableType == UP_DOWN) */
			{
				CurrentConnection = UDLookUp[Index].FirstConnection;

				if (CurrentConnection == NULL_)
				{
					printf("[X]");
				}
				else
				{
					printf("[");

					do
					{
						if (CurrentConnection->NextConnection == UDLookUp[Index].FirstConnection)
							printf("%d]", CurrentConnection->ConnectionID);
						else
							printf("%d,", CurrentConnection->ConnectionID);

						CurrentConnection = CurrentConnection->NextConnection;
					}
					while (CurrentConnection != UDLookUp[Index].FirstConnection);
				}
			}


		}
		printf("\n");
	} 
	printf("\n");
}	


/* 
*
*	  S T A T I S T I C S   P R I N T I N G   R O U T I N E S 
*
*/


/*
*	PrintIRNetStats
*
*	Prints out the statistics for the Irregular Topology.
*
*/

void	PrintIRNetStats()
{
	int	i;

}




/* 
*
*	  F I L E   I N P U T / O U T P U T   R O U T I N E S 
*
*/

/*
*	WriteNetworktoFile
*
*	Writes contents of adjacency list into specified file  
*	in the same format as input file.
*
*
*/

int WriteNetworktoFile(OutputFileName)
char	*OutputFileName;
{
	int 		i;
	FILE 		*OutputFile;
	NETNODEPTR	CurrentNetNode;

	/* open the file for writing */
	OutputFile = fopen(OutputFileName,"w");
	if(OutputFile == NULL_)
	{
		printf("ERROR: Openning File in 'WriteNetworktoFile'\n");
		return(ERROR);
	}

	/* print out file header */
	fprintf(OutputFile, "IRNUMOFNODES=%d\n", IRNUMOFNODES);
	fprintf(OutputFile, "IRNUMOFLINKS=%d\n", IRNUMOFLINKS);

	/* print out node connections */
	for (i = 0; i < IRNUMOFNODES; i++)
	{
		CurrentNetNode = MainNetNodeList[i].FirstNetNode;

		/* print out the list of VC nodes */
		while (CurrentNetNode != NULL_) 
		{
			if (i < CurrentNetNode->NodeID)
				fprintf(OutputFile, "%d-%d\n", i, CurrentNetNode->NodeID);

			CurrentNetNode = (CurrentNetNode->NextNetNode);
		}
	}

	fclose(OutputFile);

	return(NOERROR);
}


/*
*	WriteRoutingTabletoFile
*
*	Writes contents of routing table into specified file.  
*
*	'TableType' should be one of the following values:
*
*	STATIC: causes generating table in single entry per 
*		current node/dest. node pair (same format
*		as the input file when manually specifying static 
*		routing table).
*
*	ADAPTIVE: causes generating table in multiple entries per 
*		current node/dest (enclosed in brackets, separated by commas).
*		Should be same as the input file when manually specifying 
*		adaptive routing table).
*
*	UP_DOWN: causes generating table in multiple entries per 
*		current node/dest  for the Up/Down Table (enclosed in brackets, 
*		separated by commas).
*		Note this is similar, but a separate structure from the adaptive
*		routing table. 
*/

int WriteRoutingTabletoFile(OutputFileName, TableType)
char	*OutputFileName;
int	TableType;
{
	int 		i, j;
	FILE 		*OutputFile;
	CONNLISTITEMPTR	CurrentConnection;
	int		Index;

	/* open the file for writing */
	OutputFile = fopen(OutputFileName,"w");
	if(OutputFile == NULL_)
	{
		printf("ERROR: Openning File in 'WriteRoutingTabletoFile'\n");
		return(ERROR);
	}

	/* print out file header */
	fprintf(OutputFile, "TABLESIZE=%d\n", IRNUMOFNODES);


	for( i=0 ; i < IRNUMOFNODES; i++)
	{
		for( j=0 ; j < IRNUMOFNODES; j++)
		{
			Index = (i * IRNUMOFNODES) + j;

			if (TableType == STATIC)
			{
				if (i == j)
					fprintf(OutputFile, "%d ", 0);
				else
					fprintf(OutputFile, "%d ", LookUp[Index].StaticConnectionID);
			}

			else if (TableType == ADAPTIVE)
			{
				CurrentConnection = LookUp[Index].FirstConnection;

				if (CurrentConnection == NULL_)
				{
					fprintf(OutputFile, "[X]");
				}
				else
				{
					fprintf(OutputFile, "[");

					do
					{
						if (CurrentConnection->NextConnection == LookUp[Index].FirstConnection)
							fprintf(OutputFile,"%d]", CurrentConnection->ConnectionID);
						else
							fprintf(OutputFile,"%d,", CurrentConnection->ConnectionID);

						CurrentConnection = CurrentConnection->NextConnection;
					}
					while (CurrentConnection != LookUp[Index].FirstConnection);
				}
			}

			else /* if (TableType == UP_DOWN) */
			{
				CurrentConnection = UDLookUp[Index].FirstConnection;

				if (CurrentConnection == NULL_)
				{
					fprintf(OutputFile, "[X]");
				}
				else
				{
					fprintf(OutputFile, "[");

					do
					{
						if (CurrentConnection->NextConnection == UDLookUp[Index].FirstConnection)
							fprintf(OutputFile,"%d]", CurrentConnection->ConnectionID);
						else
							fprintf(OutputFile,"%d,", CurrentConnection->ConnectionID);

						CurrentConnection = CurrentConnection->NextConnection;
					}
					while (CurrentConnection != UDLookUp[Index].FirstConnection);
				}
			}

		}

		fprintf(OutputFile, "\n");
	}
	
	/* temp hack */
	for( i=0 ; i < IRNUMOFNODES; i++)
	{
		fprintf(OutputFile, "%d>%d\n", i, MainNetNodeList[i].Level);
	}

	fclose(OutputFile);

	return(NOERROR);
}


/*
*	WriteNetworktoDaVinciFile
*
*	Reads in network adjacency list and creates a visual representation
*	of the topology in daVinci format with bidirectional links and
*	NO edge labels. Nodes are labeled with Node ID.
*
*/

int WriteNetworktoDaVinciFile(char *DaVinciFileName)
{
	FILE *OutputFile;

	int node1, node2;
	int i;
	NETNODEPTR	CurrentNetNode;

	/* first pass ... create connected graph */

	OutputFile = fopen(DaVinciFileName,"w");
	if(OutputFile == NULL_)
	{
		printf("ERROR: Openning File in 'GenerateDaVinciFile'\n");
		return(ERROR);
	}

	fprintf(OutputFile, "[ \n");

	for (i = 0; i < IRNUMOFNODES; i++)
	{

		if (i == (IRNUMOFNODES - 1))
		{
			fprintf(OutputFile, "l(\"%d\",n(\"anything\",[a(\"OBJECT\",\" %d \")],[]))\n", i, i);
		}
		else
		{
			fprintf(OutputFile, "l(\"%d\",n(\"anything\",[a(\"OBJECT\",\" %d \")],", i, i);

			fprintf(OutputFile, "[ \n");

			CurrentNetNode = MainNetNodeList[i].FirstNetNode;

			while (CurrentNetNode != NULL_) 
			{
				if (CurrentNetNode->NodeID > i)
				{
					{
						fprintf(OutputFile, "e(\"anything\",[a(\"_DIR\",\"both\")],\n");

						if (CurrentNetNode->NextNetNode != NULL_)
							fprintf(OutputFile, "r(\"%d\")),\n", CurrentNetNode->NodeID);
						else
							fprintf(OutputFile, "r(\"%d\"))\n", CurrentNetNode->NodeID);

					}
				}

				CurrentNetNode = (CurrentNetNode->NextNetNode);
			}

			fprintf(OutputFile, "])), \n");
		}
	}

	fprintf(OutputFile, "] \n");
	fclose(OutputFile);

	return(NOERROR);

}


/* 
*
*	  N E T W O R K   G E N E R A T I O N   R O U T I N E S 
*
*/


/*
*	GenerateNetwork
*
*	Generates a random (connected) network topology
*	based on the following specified parameters:
*
*	- NumOfNodes
*	- NumOfLinks
*	- MinDegree
*	- MaxDegree
*	- AllowDuplicates
*
*	Note: Current algorithm biases the network nodes
*	created earlier with higher node degrees (as a result
*	of attempting to create a connected graph).
*/

int GenerateNetwork(NumOfNodes, NumOfLinks, MinDegree, MaxDegree, AllowDuplicates)
int NumOfNodes;
int NumOfLinks;
int MinDegree;
int MaxDegree;
int AllowDuplicates;
{
	int node1, node2;
	int i, Error;
	int links_added;
	int attempts;
	int MaxAttempts;
	
	/* Number of attempts at finding a node to connect to (useing random generation) */
	MaxAttempts = NumOfNodes * 10000;

	/* During the first pass, generate a simple connected graph 	*/
	/* which includes all 'NumOfNodes' nodes and 'NumOfLinks - 1'	*/
	/* links							*/

	if (NumOfNodes > MAXNODES)
	{
		printf("ERROR: Number of nodes specified by 'IRNUMOFNODES' exceeds 'MAXNODES' in 'GenerateNetwork'\n");
		return(ERROR);

	}

	if (NumOfLinks < (NumOfNodes - 1))
	{
		printf("ERROR: Number of links specified by 'IRNUMOFLINKS' insufficient to create a connected graph (in 'GenerateNetwork')\n");
		return(ERROR);

	}

	/* During the first pass, generate a simple connected graph 	*/
	/* which includes all 'NumOfNodes' nodes and 'NumOfLinks - 1'	*/
	/* links							*/

	links_added = 0;
	node1 = 0;

	for (i = 1 ; i < NumOfNodes; i++)
	{

		/* 
		printf("Adding: %d-%d\n", node1, i);
		*/

		Error = AddNetNodeToList(node1, i);

		if (Error != NOERROR)
		{
			printf("ERROR: while adding to list in 'GenerateNetwork'\n");
			return(Error);
		}

		Error = AddNetNodeToList(i, node1);

		if (Error != NOERROR)
		{
			printf("ERROR: while adding to list in 'GenerateNetwork'\n");
			return(Error);
		}

		links_added++;

		attempts = 0;
		do 
		{
			node1 = (int) (drand48() * i);
			attempts++;
		}
		while ((MainNetNodeList[node1].Degree >= MaxDegree) && (attempts < MaxAttempts));

		if (attempts >= MaxAttempts)
		{
			printf("ERROR: unable to find eligible node to connect to in 'GenerateNetwork'\n");
			return(ERROR);
		}

	}

	/* During the second pass, we attempt to satisfy the minimum 	*/
	/* degree requirements of all nodes (if not already satisfied.	*/

	for (i = 0 ; i < NumOfNodes; i++)
	{
		while(MainNetNodeList[i].Degree < MinDegree)
		{
			attempts = 0;
			do
			{
				node1 = (int) (drand48() * NumOfNodes);
				attempts++;
			}
			while (((MainNetNodeList[node1].Degree >= MaxDegree) || (MainNetNodeList[node1].Degree >= MinDegree) || (node1 == i)) &&
				(attempts < MaxAttempts));

			if (attempts >= MaxAttempts)
			{
				printf("ERROR: unable to find eligible node to connect to in 'GenerateNetwork'\n");
				return(ERROR);
			}

			/* if we are here, then it is ok --- so add the link */
			Error = AddNetNodeToList(node1, i);

			if (Error != NOERROR)
			{
				printf("ERROR: while adding to list in 'GenerateNetwork'\n");
				return(Error);
			}

			Error = AddNetNodeToList(i, node1);

			if (Error != NOERROR)
			{
				printf("ERROR: while adding to list in 'GenerateNetwork'\n");
				return(Error);
			}

			links_added++;

		}
	}

	if (links_added > NumOfLinks)
	{
		printf("ERROR: The minimum degree specified by IRMINDEGREE cannot be satisfied using only the specified number of links (in 'GenerateNetwork')\n");
		return(ERROR);
	}


	/* During the third pass, generate any additional links */
	/* required to satisfy the number of links specified.	*/

	for (i = links_added ; i < NumOfLinks; i++)
	{
		attempts = 0;
		do
		{
			node1 = (int) (drand48() * NumOfNodes);
			attempts++;
		}
		while ((MainNetNodeList[node1].Degree >= MaxDegree) && (attempts < MaxAttempts));

		if (attempts >= MaxAttempts)
		{
			printf("ERROR: unable to find eligible node to connect to in 'GenerateNetwork' (degree exceeded by all nodes)\n");
			return(ERROR);
		}

		attempts = 0;
		do
		{
			node2 = (int) (drand48() * NumOfNodes);
			attempts++;
		}
		while (((MainNetNodeList[node2].Degree >= MaxDegree) || (node1 == node2) ||
			(!AllowDuplicates && (FindEntryInAdjacencyList(node1, node2)))) &&
			(attempts < MaxAttempts));

		if (attempts >= MaxAttempts)
		{
			printf("ERROR: unable to find eligible node to connect to in 'GenerateNetwork'\n");
			return(ERROR);
		}


		Error = AddNetNodeToList(node1, node2);

		if (Error != NOERROR)
		{
			printf("ERROR: while adding to list in 'GenerateNetwork'\n");
			return(Error);
		}

		Error = AddNetNodeToList(node2, node1);

		if (Error != NOERROR)
		{
			printf("ERROR: while adding to list in 'GenerateNetwork'\n");
			return(Error);
		}

		links_added++;
	}

	return(NOERROR);
}


/*
*	GenerateNetworkFromFile
*
*	Reads in Network Topology from Input File (instead of automatically 
*	generating the topology) and creates the adjacency list representing
*	specified graph topology.
*
*	Returns the Number of Nodes and Links specified in the input file.
*
*/

int GenerateNetworkFromFile(InputFileName, NumOfNodes, NumOfLinks)
char *InputFileName;
int  *NumOfNodes;
int  *NumOfLinks;
{
	FILE *InputFile;
	int   nodes, links;
	int   i, j, k;
	int   Error;

	InputFile = fopen(InputFileName,"r");
	if(InputFile == NULL_)
	{
		printf("ERROR: Openning File in 'GenerateNetworkFromFile'\n");
		exit(0);
	}

	fscanf(InputFile, "IRNUMOFNODES=%d\n", &nodes);
	fscanf(InputFile, "IRNUMOFLINKS=%d\n", &links);

	/* Error check */
	if (nodes > links+1)
	{
		printf("ERROR: Insufficient number of links specified by 'IRNUMOFLINKS' in 'GenerateNetworkFromFile'\n");
		return(ERROR);

	}
	if (nodes > MAXNODES)
	{
		printf("ERROR: Number of nodes specified by 'IRNUMOFNODES' exceeds 'MAXNODES' in 'GenerateNetworkFromFile'\n");
		return(ERROR);

	}

	/* Print out (echo) input parameters found in file */
	/* printf("IRNUMOFNODES=%d, IRNUMOFLINKS=%d\n", nodes, links); */

	*NumOfNodes = nodes;
	*NumOfLinks = links;

	for( i=0 ; i < links; i++)
	{
		if(fscanf(InputFile, "%d-%d\n", &j, &k))
		{
			if (j == k) 
			{
				printf("ERROR: link connecting node to itself found in 'GenerateNetworkFromFile'\n");
				return(ERROR);
			}

			Error = AddNetNodeToList(j, k);

			if (Error != NOERROR)
			{
				printf("ERROR: while adding to list in 'GenerateNetworkFromFile'\n");
				return(ERROR);
			}

			Error = AddNetNodeToList(k, j);

			if (Error != NOERROR)
			{
				printf("ERROR: while adding to list in 'GenerateNetworkFromFile'\n");
				return(ERROR);
			}
		}
	} 
	
	fclose(InputFile);

	return(NOERROR);
}





/* 
*
*	  A D J A C E N C Y   L I S T   M A N A G E M E N T   R O U T I N E S 
*
*/

/*
*	AddNetNodeToList
*
*	Given a Network Node, adds it as being adjacent to a specified Node.
*
*	Returns ERROR if:
*		- an invalid VC or Circuit ID are specified.
*
*	Returns NOERROR otherwise.
*
*/

int	AddNetNodeToList(NodeID, NewNodeID)
int	NodeID;
int	NewNodeID;
{

	NETNODEPTR	CurrentNetNode, PreviousNetNode, NewNetNode;

	/*
	printf("Entering: 'AddNetNodeToList': %d-%d\n", NodeID, NewNodeID);
	*/

	/* range check array bounds */	
	if ((NodeID >= MAXNODES) || (NewNodeID >= MAXNODES))
	{
		printf("ERROR: 'NodeID' and/or 'NewNodeID' out of range in 'AddNetNodeToList'\n");
		return(ERROR);
	}

	NewNetNode = (NETNODEPTR) malloc(sizeof(NETNODE));

	/* check if memory allocation failed or not */
	if (NewNetNode == NULL_)
	{
		printf("ERROR: malloc returned NULL_ in 'AddNetNodeToList'\n");
			return(ERROR);
	}

	/* Initialize the Node */
	NewNetNode->NodeID = NewNodeID;
	NewNetNode->PhysLink = 0;
	NewNetNode->Connection = 0;
	NewNetNode->NextNetNode = NULL_;

	/* if list empty or the new node belongs at the begining of the list, insert at begining */
	if ((MainNetNodeList[NodeID].FirstNetNode == NULL_) || (MainNetNodeList[NodeID].FirstNetNode->NodeID >= NewNodeID))
	{
		NewNetNode->NextNetNode = MainNetNodeList[NodeID].FirstNetNode;
		MainNetNodeList[NodeID].FirstNetNode = NewNetNode;
	}
	/* else if it belongs elsewhere, then insert there */
	else
	{
		CurrentNetNode = MainNetNodeList[NodeID].FirstNetNode;
		PreviousNetNode = MainNetNodeList[NodeID].FirstNetNode;

		while ((CurrentNetNode != NULL_) &&
			(CurrentNetNode->NodeID < NewNodeID))
		{
			PreviousNetNode = CurrentNetNode;
			CurrentNetNode = CurrentNetNode->NextNetNode;
		}
		
		PreviousNetNode->NextNetNode = NewNetNode;
		NewNetNode->NextNetNode = CurrentNetNode;
	}

	MainNetNodeList[NodeID].Degree++;

	/*
	printf("Exiting: 'AddNetNodeToList': %d-%d\n", NodeID, NewNodeID);
	*/

	return(NOERROR);	
}



/*
*	FindEntryInAdjacencyList
*
*	Given a list (specified by ListIndex) and an EntryNodeID
*	Finds if the specified entry is in the list.
*
*	Returns TRUE if entry in the list.  Returns FALSE otherwise.
*
*/


int FindEntryInAdjacencyList(ListIndex, EntryNodeID)
int ListIndex;
int EntryNodeID;
{
	int		Found;
	NETNODEPTR	CurrentNetNode;
	
	Found = FALSE;

	CurrentNetNode = MainNetNodeList[ListIndex].FirstNetNode;

	while (CurrentNetNode != NULL_) 
	{
		if (CurrentNetNode->NodeID == EntryNodeID)
		{
			Found = TRUE;
			break;
		}

		CurrentNetNode = CurrentNetNode->NextNetNode;
	}
		
	return(Found);

}


/*
*	SetPhysicalLink
*
*	Given a adjacency list (specified by ListIndex) and an
*	Entry within the list (specified by EntryNodeID), 
*	sets the Physical Link ID of that entry.
*
*/

void SetPhysicalLink(ListIndex, EntryNodeID, PhysicalLinkID)
int ListIndex;
int EntryNodeID;
int PhysicalLinkID;
{
	NETNODEPTR	CurrentNetNode;
	
	CurrentNetNode = MainNetNodeList[ListIndex].FirstNetNode;

	while (CurrentNetNode != NULL_) 
	{
		if (CurrentNetNode->NodeID == EntryNodeID)
		{
			CurrentNetNode->PhysLink = PhysicalLinkID;
			break;
		}

		CurrentNetNode = CurrentNetNode->NextNetNode;
	}
		
}



/*
*	SetUpLinks
*
*	Traverses each node in the adjacentcy list
*	and assigns one physical link to "node pair".
*
*/

void	SetUpLinks(NumOfNodes)
int	NumOfNodes;
{
	int		i, j;
	NETNODEPTR	CurrentNetNode;
	
        j = 0;
	for (i = 0; i < NumOfNodes; i++)
	{
		CurrentNetNode = MainNetNodeList[i].FirstNetNode;

		while (CurrentNetNode != NULL_) 
		{
			if (i < CurrentNetNode->NodeID)
			{
				plinkTable[j].Lnode = i;
				plinkTable[j].Rnode = CurrentNetNode->NodeID;

				CurrentNetNode->PhysLink = j;
				/* above line the same as: 	SetPhysicalLink(i, CurrentNetNode->NodeID, j); */

				SetPhysicalLink(CurrentNetNode->NodeID, i, j);

				j++;
			}
			CurrentNetNode = CurrentNetNode->NextNetNode;
		}
	}
}



/*
*	SetConnections
*
*	Traverses the adjacency list of each node
*	and assigns logical connection IDs for
*	the physical links which connect to 
*	each adjacent node (starting from 0).
*
*/

void	SetConnections(NumOfNodes)
int	NumOfNodes;
{
	int		i, j;
	NETNODEPTR	CurrentNetNode;
	
	for (i = 0; i < NumOfNodes; i++)
	{
        	j = 0;

		CurrentNetNode = MainNetNodeList[i].FirstNetNode;

		while (CurrentNetNode != NULL_) 
		{
			CurrentNetNode->Connection = j;

			j++;

			CurrentNetNode = CurrentNetNode->NextNetNode;
		}
	}
}




/* 
*
*	  F I F O   Q U E U E   M A N A G E M E N T   R O U T I N E S 
*
*/

/*
*	AllocateQueue
*
*	Allocates memory for FIFO queue used in BFS.
*
*/

int	AllocateQueue()
{
	Queue = (int *) malloc(sizeof(int) * (IRNUMOFNODES+1));
	if (Queue == NULL_)
	{
		printf("ERROR: malloc returned NULL_ in 'AllocateQueue'\n");
		return(ERROR);
	}
	return(NOERROR);
}

/*
*	InitQueue
*
*	Initializes FIFO queue used in BFS.
*
*/

void	InitQueue()
{
	int i;

	for (i = 0; i < IRNUMOFNODES+1 ; i++)
		Queue[i] = 0;

	HeadOfQueue = 0;
	TailOfQueue = HeadOfQueue;
}


/*
*	AddToQueue
*
*	Adds specified Entry in to FIFO queue used in BFS.
*
*/

int	AddToQueue(NewQueueEntry)
int	NewQueueEntry;
{
	int i;

	if (TailOfQueue == IRNUMOFNODES)
	{
		printf("ERROR: Queue Full in 'AddToQueue'\n");
		return(ERROR);
	}
	else
	{
		Queue[TailOfQueue] = NewQueueEntry;
		TailOfQueue++;
		return(NOERROR);
	}
}


/*
*	RemoveFromQueue
*
*	Removes and returns item from FIFO Queue used in BFS.
*	Should be called only if Queue is Not Empty (call 'QueueEmpty' first).
*
*/

int	RemoveFromQueue()
{
	int	Entry;

	Entry = Queue[HeadOfQueue];
	HeadOfQueue++;
	return(Entry);
}


/*
*	QueueEmpty
*
*	Returns TRUE if FIFO  Queue is empty (used in BFS).
*	Returns FALSE otherwise.
*
*/

int	QueueEmpty()
{
	if (HeadOfQueue == TailOfQueue)
		return(TRUE);
	else
		return(FALSE);
}



/* 
*
*	  R O U T I N G   T A B L E   M A N A G E M E N T   R O U T I N E S 
*
*/

/*
*	AllocateRoutingTable
*
*	Allocates memory for 2-Dimensional routing table of size IRNUMOFNODES^2.
*
*/

int	AllocateRoutingTable()
{
	int	i;

	LookUp = (TABLEENTRYPTR) malloc(sizeof(TABLEENTRY) * IRNUMOFNODES * IRNUMOFNODES);

	/* check if memory allocation failed or not */
	if (LookUp == NULL_)
	{
		printf("ERROR: malloc returned NULL_ in 'AllocateRoutingTable'\n");
			return(ERROR);
	}

	return(NOERROR);
}


/*
*	FreeRoutingTable
*
*	Frees all of the memory allocated for the routing table.
*
*/

void	FreeRoutingTable()
{
	int		i, j;

	for( i=0 ; i < IRNUMOFNODES; i++)
	{
		for( j=0 ; j < IRNUMOFNODES; j++)
		{
			FreeRoutingTableList(i, j);
		}
	}
	free(LookUp);
}


/*
*	FreeRoutingTableList
*
*	Frees the list of connections associated with a routing table entry.
*
*/

void	FreeRoutingTableList(i, j)
int	i, j;
{
        int 		Index;
	CONNLISTITEMPTR	FirstConnection, CurrentConnection, NextConnection;	

	Index = (i * IRNUMOFNODES) + j;

	FirstConnection = LookUp[Index].FirstConnection;
	CurrentConnection = LookUp[Index].FirstConnection;

	if ((FirstConnection != NULL_) && (FirstConnection->NextConnection == FirstConnection))
		free(FirstConnection);

	else
	{
		while ((CurrentConnection != NULL_) && (CurrentConnection != FirstConnection))
		{
			NextConnection = CurrentConnection->NextConnection;

			free(CurrentConnection);

			CurrentConnection = NextConnection;
		}

		if (CurrentConnection != NULL_)
			free(CurrentConnection);
	}
}


/*
*	InitializeRoutingTable
*
*	Initializes 2-Dimensional routing table.
*
*/

void	InitializeRoutingTable()
{
	int   i, j, Index;


	for( i=0 ; i < IRNUMOFNODES; i++)
		for( j=0 ; j < IRNUMOFNODES; j++)
		{
			Index = (i * IRNUMOFNODES) + j;
			LookUp[Index].MinDistance = 0;
			LookUp[Index].NumOfConnections = 0;
			LookUp[Index].StaticConnectionID = -9;
			LookUp[Index].FirstConnection = NULL_;
		}
}	


/*
*	AddConnectionToList
*
*	Adds a specified connection (NewConnID & NewDistance) to
*	the routing table entry list specified via CurrentNode & DestNod
*
*/

int	AddConnectionToList(CurrentNode, DestNode, NewConnID, NewDistance)
int	CurrentNode;
int	DestNode;
int	NewConnID;
int	NewDistance;
{
        int 		Index;
	CONNLISTITEMPTR	CurrentConnection, PreviousConnection, NewConnection;	

	Index = (CurrentNode * IRNUMOFNODES) + DestNode;

	if ((LookUp[Index].FirstConnection == NULL_) || (LookUp[Index].FirstConnection->ConnectionID > NewConnID))
	{
		NewConnection = (CONNLISTITEMPTR) malloc(sizeof(CONNLISTITEM));

		/* check if memory allocation failed or not */
		if (NewConnection == NULL_)
		{
			printf("ERROR: malloc returned NULL_ in 'AddConnectionToList'\n");
			return(ERROR);
		}

		/* Set up values of newly created connection list item */
		NewConnection->ConnectionID = NewConnID;
		NewConnection->Distance = NewDistance; 			/* not used for minimal algorithm */

		/* If it's the first item to be added to the list, make it point to itself */
		if (LookUp[Index].FirstConnection == NULL_) 
		{
			NewConnection->NextConnection = NewConnection;
		}
		/* Otherwise, find the last item on the list and make it point to the newly added item at the head of list */
		else
		{
			/* Find the last item on the list --- and make it point to the newly added first item */
			CurrentConnection = LookUp[Index].FirstConnection;

			while (CurrentConnection->NextConnection != LookUp[Index].FirstConnection)
			{
				CurrentConnection = CurrentConnection->NextConnection;
			}

			CurrentConnection->NextConnection = NewConnection;

			/* Also have the newly added first item point to the second item on the list */
			NewConnection->NextConnection = LookUp[Index].FirstConnection;
		}

		/* In either case, have the first item be the new item */
		LookUp[Index].FirstConnection = NewConnection;
	}


	else
	{
		CurrentConnection = LookUp[Index].FirstConnection;
		PreviousConnection = LookUp[Index].FirstConnection;

		while ((CurrentConnection->NextConnection != LookUp[Index].FirstConnection) &&
			(CurrentConnection->ConnectionID < NewConnID))
		{
			PreviousConnection = CurrentConnection;
			CurrentConnection = CurrentConnection->NextConnection;
		}

		/* simply ignore request to add if the connection already exists */		
		if (CurrentConnection->ConnectionID == NewConnID)
			return(NOERROR);

		/* otherwise, allocate memory and add to linked list */		
		NewConnection = (CONNLISTITEMPTR) malloc(sizeof(CONNLISTITEM));

		/* check if memory allocation failed or not */
		if (NewConnection == NULL_)
		{
			printf("ERROR: malloc returned NULL_ in 'AddConnectionToList'\n");
			return(ERROR);
		}

		NewConnection->ConnectionID = NewConnID;
		NewConnection->Distance = NewDistance; 			/* not used for minimal algorithm */

		/* If it goes as the new last item on the list  (i.e. existing last element is smaller than new element) */		
		if (CurrentConnection->ConnectionID < NewConnID)
		{
			NewConnection->NextConnection = CurrentConnection->NextConnection;
			CurrentConnection->NextConnection = NewConnection;
		}
		/* If it goes right before existing last item on the list (i.e. new element is smaller than existing last element) */		
		else
		{
			PreviousConnection->NextConnection = NewConnection;
			NewConnection->NextConnection = CurrentConnection;
		}
	}

	LookUp[Index].NumOfConnections++;

	return(NOERROR);	

}	


/*
*	AddConnectionListToList
*
*	Adds all of the elements in the connection list of table entry 
*	specified via (FromCurrentNode, FromDestNode) to the connection
*	list of table entry specified via (ToCurrentNode, ToDestNode)
*/

int	AddConnectionListToList(FromCurrentNode, FromDestNode, ToCurrentNode, ToDestNode)
int	FromCurrentNode;
int	FromDestNode;
int	ToCurrentNode;
int	ToDestNode;
{
        int 		Index;
        int 		Error;
	CONNLISTITEMPTR	CurrentConnection;	

	Index = (FromCurrentNode * IRNUMOFNODES) + FromDestNode;

	CurrentConnection = LookUp[Index].FirstConnection;

	while (CurrentConnection != NULL_)
	{
		Error = AddConnectionToList(ToCurrentNode, ToDestNode, 
			CurrentConnection->ConnectionID, CurrentConnection->Distance);

		if (Error != NOERROR)
		{
			printf("ERROR: malloc returned NULL_ in 'AddConnectionToList'\n");
			return(ERROR);
		}

		CurrentConnection = CurrentConnection->NextConnection;

		/* break out of loop when we reach the end --- before circling again */
		if (CurrentConnection == LookUp[Index].FirstConnection)
			break;
	}

	return(NOERROR);
}	



/*
*	GenerateStaticRoutingTableFromFile
*
*	Reads specified file containing a static (2D) routing
*	table (single entry per current/detination node pair)
*	and creates a static routing table.
*
*/

int	GenerateStaticRoutingTableFromFile(InputFileName, NumOfEntries)
char	*InputFileName;
int	*NumOfEntries;
{
	FILE *InputFile;
	int   entries, connection;
	int   i, j;
	int   Error;
	int   Index;

	InputFile = fopen(InputFileName,"r");
	if(InputFile == NULL_)
	{
		printf("ERROR: Openning File in 'GenerateStaticRoutingTableFromFile'\n");
		exit(0);
	}

	fscanf(InputFile, "TABLESIZE=%d\n", &entries);

	*NumOfEntries = entries;

	for( i=0 ; i < entries; i++)
	{
		for( j=0 ; j < entries; j++)
		{
	      		if(fscanf(InputFile,"%d",&connection)==1)
			{
				Index = (i * entries) + j;

				LookUp[Index].NumOfConnections = 1;

				LookUp[Index].FirstConnection =  (CONNLISTITEMPTR) malloc(sizeof(CONNLISTITEM));
				if (LookUp[Index].FirstConnection == NULL_)
				{
					printf("ERROR: malloc returned NULL_ in 'GenerateStaticRoutingTableFromFile'\n");
					return(ERROR);
				}

				LookUp[Index].FirstConnection->ConnectionID =  connection;
				LookUp[Index].FirstConnection->Distance =  0;
				LookUp[Index].FirstConnection->NextConnection =  NULL_;
			}
		}
		fscanf(InputFile,"\n");
	} 
	
	fclose(InputFile);

	return(NOERROR);
}	



/*
*	DetermineStaticRoutingConnections
*
*	Simply selects one of the routing options (determined
*	by the adaptive routing table generation) for use 
*	in static routing. The selection is random.
*	Once the routing entry is selected, the list is
*	freed (no longer needed).
*
*/

void	DetermineStaticRoutingConnections()
{
	int i, j, k;
	TABLEENTRYPTR TableEntry;
	int NumOfPLs, randomPL;
        CONNLISTITEMPTR	CurrentConnection;

	for (i = 0; i < IRNUMOFNODES; i++)
	{
		for (j = 0; j < IRNUMOFNODES; j++)
		{
			if (i != j)
			{
       				 /* Look up in the routing tbale to get a list of eligible PLs ("connections") */
        			TableEntry = (TABLEENTRYPTR) AdaptiveTableLookUp(i, j);

				/* The total number of Pls available for use */
      				NumOfPLs = TableEntry->NumOfConnections;

      				/* Randomly select the first PL to consider */
      				randomPL = (int) (drand48() * NumOfPLs);
      				CurrentConnection = TableEntry->FirstConnection;

      				for(k=0 ; k < randomPL ; k++)
          				CurrentConnection = CurrentConnection->NextConnection;

				LookUp[(i * IRNUMOFNODES) + j].StaticConnectionID = CurrentConnection->ConnectionID;

				FreeRoutingTableList(i, j);
			}
		}
	}
}

/*
*	StaticTableLookUp
*
*	Simply returns the pre-determined item from 
*	the connection list of the table entry specified via 
*	(CurrentNode, DestNode). This is intended
*	to be used for static routing functions.
*/

int	StaticTableLookUp(CurrentNode, DestNode)
int	CurrentNode;
int	DestNode;
{
	return(LookUp[(CurrentNode * IRNUMOFNODES) + DestNode].StaticConnectionID);
}	


/*
*	GenerateMinTFARoutingTable
*
*	Generates Routing Table for Minimal True Fully Adaptive Routing.
*	Breadth First Search is used (with enhancements) to find all
*	the connections (links) which begin shortest paths between all node pairs.
*	These connections are placed in the connection lists of the
*	corresponding table entries.
*	
*/

int	GenerateMinTFARoutingTable()
{
	int		i, j;
	int		Error;
	int		CurrentNodeID;
	NETNODEPTR	CurrentNetNode;
	int		FirstPass;

	Error = AllocateQueue();
	if (Error != NOERROR)
	{
		printf("ERROR: Adding to Queue in 'GenerateMinTFARoutingTable'\n");
		return(ERROR);
	}

	for (i = 0; i < IRNUMOFNODES; i++)
	{
		InitColor();
		InitQueue();
		FirstPass = TRUE;

		MainNetNodeList[i].Color = GRAY;
		LookUp[(i * IRNUMOFNODES) + i].MinDistance = 0;

		Error = AddToQueue(i);
		if (Error != NOERROR)
		{
			printf("ERROR: Adding to Queue in 'GenerateMinTFARoutingTable'\n");
			return(ERROR);
		}

		while(!QueueEmpty())
		{
			CurrentNodeID = RemoveFromQueue();

			CurrentNetNode = MainNetNodeList[CurrentNodeID].FirstNetNode;

			while (CurrentNetNode != NULL_)
			{
				if (MainNetNodeList[CurrentNetNode->NodeID].Color == WHITE)
				{
					MainNetNodeList[CurrentNetNode->NodeID].Color = GRAY;
					LookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance = 
						LookUp[(i * IRNUMOFNODES) + CurrentNodeID].MinDistance + 1; 

					if (FirstPass)
					{
						Error = AddConnectionToList(i, CurrentNetNode->NodeID, CurrentNetNode->Connection, 0);
						if (Error != NOERROR)
						{
							printf("ERROR: Adding Connection to List in 'GenerateMinTFARoutingTable'\n");
							return(ERROR);
						}
					}
					else
					{
						Error = AddConnectionListToList(i, CurrentNodeID, i, CurrentNetNode->NodeID);
						if (Error != NOERROR)
						{
							printf("ERROR: Adding Connection to List in 'GenerateMinTFARoutingTable'\n");
							return(ERROR);
						}
					}
		
					Error = AddToQueue(CurrentNetNode->NodeID);
					if (Error != NOERROR)
					{
						printf("ERROR: Adding to Queue in 'GenerateMinTFARoutingTable'\n");
						return(ERROR);
					}
				}
				else if (MainNetNodeList[CurrentNetNode->NodeID].Color == GRAY)
				{
					if (LookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance <  
						(LookUp[(i * IRNUMOFNODES) + CurrentNodeID].MinDistance + 1))
					{
						/* ignore since a shorter path has already been found */
					}
					else if (LookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance ==  
						(LookUp[(i * IRNUMOFNODES) + CurrentNodeID].MinDistance + 1))
					{

						if (FirstPass)
						{
							Error = AddConnectionToList(i, CurrentNetNode->NodeID, CurrentNetNode->Connection, 0);
							if (Error != NOERROR)
							{
								printf("ERROR: Adding Connection to List in 'GenerateMinTFARoutingTable'\n");
								return(ERROR);
							}
						}
						else
						{
							Error = AddConnectionListToList(i, CurrentNodeID, i, CurrentNetNode->NodeID);
							if (Error != NOERROR)
							{
								printf("ERROR: Adding Connection to List in 'GenerateMinTFARoutingTable'\n");
								return(ERROR);
							}
						}
					}
					else
					/* if (LookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance >  
						(LookUp[(i * IRNUMOFNODES) + CurrentNodeID].MinDistance + 1))
					*/
					{
						/* this case should never happen in BFS --- so report and return error */
						printf("ERROR: Found a shorter path unexpectedly in 'GenerateMinTFARoutingTable'\n");
						return(ERROR);

					}
				}

				CurrentNetNode = CurrentNetNode->NextNetNode;
			}
			
			MainNetNodeList[CurrentNodeID].Color = BLACK;
			FirstPass=FALSE;
		}
	}

	free(Queue);
	return(NOERROR);
}	


/*
*	AdaptiveTableLookUp
*
*	Returns the list of connection (a circular linked 
*	list) of the table entry specified via 
*	(CurrentNode, DestNode). This is intended
*	to be used for adaptive routing functions.
*
*/

TABLEENTRYPTR	AdaptiveTableLookUp(CurrentNode, DestNode)
int	CurrentNode;
int	DestNode;
{
	return((TABLEENTRYPTR) &LookUp[(CurrentNode * IRNUMOFNODES) + DestNode]);
}	




/* 
*
*	  D A T A   S E R V I C E S   R O U T I N E S 
*
*/

/*
*	GetNodeDegree
*
*	Simply returns the node degree of the specified node.
*/

int GetNodeDegree(NodeID)
int NodeID;
{
	return(MainNetNodeList[NodeID].Degree);
}


/*
*	GetDistance
*
*	Simply returns the routing distance between the nodes.
*/

int GetDistance(SourceNodeID, DestNodeID)
int SourceNodeID;
int DestNodeID;
{
	return(LookUp[(SourceNodeID * IRNUMOFNODES) + DestNodeID].MinDistance);
}



/*
*	GetNextHotSpot
*
*	Determines root of MST (the node with
*	the lowest average internode distance.
*	Use highest node degree as a tie-breaker).
*/

int	GetNextHotSpot(UpperBoundNodeID)
int	UpperBoundNodeID;
{
	int	i;
	int	SavedNodeID;
	int 	FoundOne;
	
	FoundOne = FALSE;
	for( i=0 ; i < IRNUMOFNODES; i++)
	{
		if ((!MainNetNodeList[i].HotSpot) && 
			(MainNetNodeList[i].Degree <= MainNetNodeList[UpperBoundNodeID].Degree))
		{
			if (!FoundOne)
			{
				SavedNodeID = i;
				FoundOne = TRUE;
			}

			else if (MainNetNodeList[SavedNodeID].Degree < MainNetNodeList[i].Degree)
			{
				SavedNodeID = i;
			}
		}
	}

	MainNetNodeList[SavedNodeID].HotSpot = TRUE;
	return(SavedNodeID);
}	




/* 
*
*	  U P / D O W N   R O U T I N E S 
*
*/


/*
*	DetermineRootOfMST
*
*	Determines root of MST (the node with
*	the lowest average internode distance.
*	Use highest node degree as a tie-breaker).
*/

int	DetermineRootOfMST()
{
	int	i, j;
	int	SumOfDistances;
	int	MinSum;
	int	RootNodeID;


	for( i=0 ; i < IRNUMOFNODES; i++)
	{
		SumOfDistances = 0;

		for( j=0 ; j < IRNUMOFNODES; j++)
		{
			if (i != j)
				SumOfDistances += LookUp[(i * IRNUMOFNODES) + j].MinDistance;

		}

		if (i == 0)
		{
			MinSum = SumOfDistances;
			RootNodeID = i;
		}
		else if ((SumOfDistances < MinSum) ||
			 ((SumOfDistances == MinSum) && (MainNetNodeList[i].Degree > MainNetNodeList[RootNodeID].Degree)))
		{
			MinSum = SumOfDistances;
			RootNodeID = i;
		}

	}

	return(RootNodeID);
}	






/*
*	DefineMSTLevels
*
*	Given a Root Node, defines a Minimum Spanning Tree on the 
*	Network Graph (by specifying Levels of other nodes).
*	
*/

int	DefineMSTLevels(RootNodeID)
int	RootNodeID;
{
	int		Error;
	int		CurrentNodeID;
	NETNODEPTR	CurrentNetNode;


	Error = AllocateQueue();
	if (Error != NOERROR)
	{
		printf("ERROR: Adding to Queue in 'DefineMSTLevels'\n");
		return(ERROR);
	}

	InitColor();
	InitQueue();

	MainNetNodeList[RootNodeID].Color = GRAY;
	MainNetNodeList[RootNodeID].Level = 0;

	Error = AddToQueue(RootNodeID);
	if (Error != NOERROR)
	{
		printf("ERROR: Adding to Queue in 'DefineMSTLevels'\n");
		return(ERROR);
	}

	while(!QueueEmpty())
	{
		CurrentNodeID = RemoveFromQueue();

		CurrentNetNode = MainNetNodeList[CurrentNodeID].FirstNetNode;

		while (CurrentNetNode != NULL_)
		{
			if (MainNetNodeList[CurrentNetNode->NodeID].Color == WHITE)
			{
				MainNetNodeList[CurrentNetNode->NodeID].Color = GRAY;
				MainNetNodeList[CurrentNetNode->NodeID].Level = MainNetNodeList[CurrentNodeID].Level + 1;

				Error = AddToQueue(CurrentNetNode->NodeID);
				if (Error != NOERROR)
				{
					printf("ERROR: Adding to Queue in 'GenerateMinTFARoutingTable'\n");
					return(ERROR);
				}
			}

			CurrentNetNode = CurrentNetNode->NextNetNode;			
		}
	}

	free(Queue);
	return(NOERROR);
}	







/* 
*
*	  U P / D O W N   F I F O   Q U E U E   M A N A G E M E N T   R O U T I N E S 
*
*/

/*
*	AllocateUDQueue
*
*	Allocates memory for Up/Down FIFO queue used in BFS.
*
*/

int	AllocateUDQueue()
{
	UDQueue = (UDQUEUEENTRYPTR) malloc(sizeof(UDQUEUEENTRY) * (MAXUDQUEUE+1));
	if (UDQueue == NULL_)
	{
		printf("ERROR: malloc returned NULL_ in 'AllocateUDQueue'\n");
		return(ERROR);
	}
	return(NOERROR);
}

/*
*	InitUDQueue
*
*	Initializes Up/Down FIFO queue used in BFS.
*
*/

void	InitUDQueue()
{
	int i;

	for (i = 0; i < MAXUDQUEUE+1 ; i++)
	{
		UDQueue[i].NodeID = 0;
		UDQueue[i].Direction = 0;
		UDQueue[i].ConnectionID = 0;
		UDQueue[i].InitialDirection = 0;
		UDQueue[i].Distance = 0;
	}

	HeadOfUDQueue = 0;
	TailOfUDQueue = HeadOfUDQueue;
}


/*
*	AddToUDQueue
*
*	Adds specified Entry in to Up/Down FIFO queue used in BFS.
*
*/

int	AddToUDQueue(NodeID, Direction, ConnectionID, InitialDirection, Distance)
int	NodeID;
int	Direction;
int	ConnectionID;
int	InitialDirection;
int	Distance;
{
	if (TailOfUDQueue == MAXUDQUEUE)
	{
		printf("ERROR: UDQueue Full in 'AddToUDQueue'\n");
		return(ERROR);
	}
	else
	{
		UDQueue[TailOfUDQueue].NodeID = NodeID;
		UDQueue[TailOfUDQueue].Direction = Direction;
		UDQueue[TailOfUDQueue].ConnectionID = ConnectionID;
		UDQueue[TailOfUDQueue].InitialDirection = InitialDirection;
		UDQueue[TailOfUDQueue].Distance = Distance;
		TailOfUDQueue++;
		return(NOERROR);
	}
}


/*
*	RemoveFromUDQueue
*
*	Removes and returns item from Up/Down FIFO queue used in BFS.
*	Should be called only if Up/Down Queue is Not Empty (call 'UDQueueEmpty' first).
*
*/

void	RemoveFromUDQueue(NodeID, Direction, ConnectionID, InitialDirection, Distance)
int	*NodeID;
int	*Direction;
int	*ConnectionID;
int	*InitialDirection;
int	*Distance;
{
	int	Entry;

	*NodeID = UDQueue[HeadOfUDQueue].NodeID;
	*Direction = UDQueue[HeadOfUDQueue].Direction;
	*ConnectionID = UDQueue[HeadOfUDQueue].ConnectionID;
	*InitialDirection = UDQueue[HeadOfUDQueue].InitialDirection;
	*Distance = UDQueue[HeadOfUDQueue].Distance;
	HeadOfUDQueue++;
}


/*
*	UDQueueEmpty
*
*	Returns TRUE if FIFO  UDQueue is empty (used in BFS).
*	Returns FALSE otherwise.
*
*/

int	UDQueueEmpty()
{
	if (HeadOfUDQueue == TailOfUDQueue)
		return(TRUE);
	else
		return(FALSE);
}



/* 
*
*	   U P / D O W N   R O U T I N G   T A B L E   M A N A G E M E N T   R O U T I N E S 
*
*/

/*
*	AllocateUDRoutingTable
*
*	Allocates memory for 2-Dimensional (Up/Down) routing table of size IRNUMOFNODES^2.
*
*/

int	AllocateUDRoutingTable()
{
	int	i;

	UDLookUp = (TABLEENTRYPTR) malloc(sizeof(TABLEENTRY) * IRNUMOFNODES * IRNUMOFNODES);

	/* check if memory allocation failed or not */
	if (UDLookUp == NULL_)
	{
		printf("ERROR: malloc returned NULL_ in 'AllocateUDRoutingTable'\n");
			return(ERROR);
	}

	return(NOERROR);
}


/*
*	FreeUDRoutingTable
*
*	Frees all of the memory allocated for the (Up/Down) routing table.
*
*/

void	FreeUDRoutingTable()
{
	int		i, j;

	for( i=0 ; i < IRNUMOFNODES; i++)
	{
		for( j=0 ; j < IRNUMOFNODES; j++)
		{
			FreeUDRoutingTableList(i, j);
		}
	}
	free(UDLookUp);
}


/*
*	FreeUDRoutingTableList
*
*	Frees the list of connections associated with a (Up/Down) routing table entry.
*
*/

void	FreeUDRoutingTableList(i, j)
int	i;
int	j;
{
        int 		Index;
	CONNLISTITEMPTR	FirstConnection, CurrentConnection, NextConnection;	

	Index = (i * IRNUMOFNODES) + j;

	FirstConnection = UDLookUp[Index].FirstConnection;
	CurrentConnection = UDLookUp[Index].FirstConnection;

	if ((FirstConnection != NULL_) && (FirstConnection->NextConnection == FirstConnection))
		free(FirstConnection);

	else
	{
		while ((CurrentConnection != NULL_) && (CurrentConnection != FirstConnection))
		{
			NextConnection = CurrentConnection->NextConnection;

			free(CurrentConnection);

			CurrentConnection = NextConnection;
		}

		if (CurrentConnection != NULL_)
			free(CurrentConnection);
	}
}


/*
*	InitializeUDRoutingTable
*
*	Initializes 2-Dimensional (Up/Down) routing table.
*
*/

void	InitializeUDRoutingTable()
{
	int   i, j, Index;


	for( i=0 ; i < IRNUMOFNODES; i++)
		for( j=0 ; j < IRNUMOFNODES; j++)
		{
			Index = (i * IRNUMOFNODES) + j;
			UDLookUp[Index].MinDistance = 0;
			UDLookUp[Index].NumOfConnections = 0;
			UDLookUp[Index].StaticConnectionID = -9;  /* Not needed for Up/Down */
			UDLookUp[Index].FirstConnection = NULL_;
		}
}	


/*
*	UDAddConnectionToList
*
*	Adds a specified connection (NewConnID & NewDistance) to
*	the routing table entry list specified via CurrentNode & DestNod
*
*/

int	UDAddConnectionToList(CurrentNode, DestNode, NewConnID, NewDistance)
int	CurrentNode;
int	DestNode;
int	NewConnID;
int	NewDistance;
{
        int 		Index;
	CONNLISTITEMPTR	CurrentConnection, PreviousConnection, NewConnection;	

	Index = (CurrentNode * IRNUMOFNODES) + DestNode;

	if ((UDLookUp[Index].FirstConnection == NULL_) || (UDLookUp[Index].FirstConnection->ConnectionID > NewConnID))
	{
		NewConnection = (CONNLISTITEMPTR) malloc(sizeof(CONNLISTITEM));

		/* check if memory allocation failed or not */
		if (NewConnection == NULL_)
		{
			printf("ERROR: malloc returned NULL_ in 'UDAddConnectionToList'\n");
			return(ERROR);
		}

		/* Set up values of newly created connection list item */
		NewConnection->ConnectionID = NewConnID;
		NewConnection->Distance = NewDistance; 			/* not used for minimal algorithm */

		/* If it's the first item to be added to the list, make it point to itself */
		if (UDLookUp[Index].FirstConnection == NULL_) 
		{
			NewConnection->NextConnection = NewConnection;
		}
		/* Otherwise, find the last item on the list and make it point to the newly added item at the head of list */
		else
		{
			/* Find the last item on the list --- and make it point to the newly added first item */
			CurrentConnection = UDLookUp[Index].FirstConnection;

			while (CurrentConnection->NextConnection != UDLookUp[Index].FirstConnection)
			{
				CurrentConnection = CurrentConnection->NextConnection;
			}

			CurrentConnection->NextConnection = NewConnection;

			/* Also have the newly added first item point to the second item on the list */
			NewConnection->NextConnection = UDLookUp[Index].FirstConnection;
		}

		/* In either case, have the first item be the new item */
		UDLookUp[Index].FirstConnection = NewConnection;
	}


	else
	{
		CurrentConnection = UDLookUp[Index].FirstConnection;
		PreviousConnection = UDLookUp[Index].FirstConnection;

		while ((CurrentConnection->NextConnection != UDLookUp[Index].FirstConnection) &&
			(CurrentConnection->ConnectionID < NewConnID))
		{
			PreviousConnection = CurrentConnection;
			CurrentConnection = CurrentConnection->NextConnection;
		}

		/* simply ignore request to add if the connection already exists */		
		if (CurrentConnection->ConnectionID == NewConnID)
			return(NOERROR);

		/* otherwise, allocate memory and add to linked list */		
		NewConnection = (CONNLISTITEMPTR) malloc(sizeof(CONNLISTITEM));

		/* check if memory allocation failed or not */
		if (NewConnection == NULL_)
		{
			printf("ERROR: malloc returned NULL_ in 'UDAddConnectionToList'\n");
			return(ERROR);
		}

		NewConnection->ConnectionID = NewConnID;
		NewConnection->Distance = NewDistance; 			/* not used for minimal algorithm */

		/* If it goes as the new last item on the list  (i.e. existing last element is smaller than new element) */		
		if (CurrentConnection->ConnectionID < NewConnID)
		{
			NewConnection->NextConnection = CurrentConnection->NextConnection;
			CurrentConnection->NextConnection = NewConnection;
		}
		/* If it goes right before existing last item on the list (i.e. new element is smaller than existing last element) */		
		else
		{
			PreviousConnection->NextConnection = NewConnection;
			NewConnection->NextConnection = CurrentConnection;
		}
	}

	UDLookUp[Index].NumOfConnections++;

	return(NOERROR);	

}	


/*
*	UDAddConnectionListToList
*
*	Adds all of the elements in the connection list of table entry 
*	specified via (FromCurrentNode, FromDestNode) to the connection
*	list of table entry specified via (ToCurrentNode, ToDestNode)
*/

int	UDAddConnectionListToList(FromCurrentNode, FromDestNode, ToCurrentNode, ToDestNode)
int	FromCurrentNode;
int	FromDestNode;
int	ToCurrentNode;
int	ToDestNode;
{
        int 		Index;
        int 		Error;
	CONNLISTITEMPTR	CurrentConnection;	

	Index = (FromCurrentNode * IRNUMOFNODES) + FromDestNode;

	CurrentConnection = UDLookUp[Index].FirstConnection;

	while (CurrentConnection != NULL_)
	{
		Error = UDAddConnectionToList(ToCurrentNode, ToDestNode, 
			CurrentConnection->ConnectionID, CurrentConnection->Distance);

		if (Error != NOERROR)
		{
			printf("ERROR: malloc returned NULL_ in 'UDAddConnectionToList'\n");
			return(ERROR);
		}

		CurrentConnection = CurrentConnection->NextConnection;

		/* break out of loop when we reach the end --- before circling again */
		if (CurrentConnection == UDLookUp[Index].FirstConnection)
			break;
	}

	return(NOERROR);
}	




/*
*	DetermineDirection
*
*	Determines direction of link based on the reltive location (level)
*	of two nodes within the minimum spanning tree (also uses the 
*	Node IDs --- if the two nodes are at the same level).
*	
*/

int	DetermineDirection(CurrentNodeID, NextNodeID)
int	CurrentNodeID;
int	NextNodeID;
{
	if (MainNetNodeList[CurrentNodeID].Level < MainNetNodeList[NextNodeID].Level)
		return(DOWN);

	else if (MainNetNodeList[CurrentNodeID].Level > MainNetNodeList[NextNodeID].Level)
		return(UP);

	else /* if (MainNetNodeList[CurrentNodeID].Level == MainNetNodeList[NextNodeID].Level) */
	{
		if (CurrentNodeID < NextNodeID)
			return(DOWN);

		else /* (CurrentNodeID > NextNodeID) */
			return(UP);
	}
}



/*
*	DetermineLinkEligibility
*
*	Determines whether a link is eligible to be used
*	based on Up/Down rules.
*	
*/

int	DetermineLinkEligibility(PreviousDirection, NewDirection)
int	PreviousDirection;
int	NewDirection;
{
	if ((PreviousDirection == UNDEFINED_DIRECTION) || 
	    (PreviousDirection == UP) ||
	    ((PreviousDirection == DOWN) && (NewDirection == DOWN)))
		return(TRUE);

	/* else */
	return(FALSE);
}



/*
*	GenerateUpDownRoutingTable
*
*	Generates Routing Table for Up/Down Routing.
*	Breadth First Search is used (with enhancements) to find all
*	the connections (links) which begin shortest allowed paths based
*	'Up/Down' rules between all node pairs (hence, some are non-minimal paths).
*	These connections are placed in the connection lists of the
*	corresponding table entries.
*
*	Color code:	white - no route
*			gray - some paths that initially goes up reach that node
*			darkgray - some paths that initially goes down reach that node
*			black - all minimal paths computed for that node
*	
*/

int	GenerateUpDownRoutingTable()
{
	int		i, j;
	int		Error;
	int		CurrentNodeID;
	int		CurrentDirection;
	NETNODEPTR	CurrentNetNode;
	int		FirstPass;
	int		NewDirection;
	int		LinkEligible;
	int		FirstConnectionID;
	int		InitialDirection;
	int		CurrentDistance;

	Error = AllocateUDQueue();
	if (Error != NOERROR)
	{
		printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
		return(ERROR);
	}

	for (i = 0; i < IRNUMOFNODES; i++)
	{
		InitColor();
		InitUDQueue();

		MainNetNodeList[i].Color = BLACK;
		UDLookUp[(i * IRNUMOFNODES) + i].MinDistance = 0;

/*		Error = AddToUDQueue(i, UNDEFINED_DIRECTION, UNDEFINED_CONNECTION);
		if (Error != NOERROR)
		{
			printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
			return(ERROR);
		}
		*/
		
		/* First Pass: Add all neighboring nodes */
		CurrentNetNode = MainNetNodeList[i].FirstNetNode;
		while (CurrentNetNode != NULL_)
		{
			if (i == CurrentNetNode->NodeID) /* ignore self loops */
			{
				CurrentNetNode = CurrentNetNode->NextNetNode;
				continue;
			}

			NewDirection = DetermineDirection(i, CurrentNetNode->NodeID);
			if (NewDirection == DOWN)
				MainNetNodeList[CurrentNetNode->NodeID].Color = DARKGRAY;
			else
				if (MainNetNodeList[CurrentNetNode->NodeID].Color == WHITE)
					MainNetNodeList[CurrentNetNode->NodeID].Color = GRAY;

			UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance = 1;

			Error = UDAddConnectionToList(i, CurrentNetNode->NodeID, CurrentNetNode->Connection, 1);
			if (Error != NOERROR)
			{
				printf("ERROR: Adding Connection to List in 'GenerateUpDownRoutingTable'\n");
				return(ERROR);
			}

			Error = AddToUDQueue(CurrentNetNode->NodeID, NewDirection, CurrentNetNode->Connection, NewDirection, 1);
			if (Error != NOERROR)
			{
				printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
				return(ERROR);
			}

			CurrentNetNode = CurrentNetNode->NextNetNode;
		}

		/* Second Pass: Process all Queued Connections */
		while(!UDQueueEmpty())
		{
			RemoveFromUDQueue(&CurrentNodeID, &CurrentDirection, &FirstConnectionID, &InitialDirection, &CurrentDistance);

			CurrentNetNode = MainNetNodeList[CurrentNodeID].FirstNetNode;

			while (CurrentNetNode != NULL_)
			{
				NewDirection = DetermineDirection(CurrentNodeID, CurrentNetNode->NodeID);

				LinkEligible = DetermineLinkEligibility(CurrentDirection, NewDirection);

				if ((LinkEligible) && (MainNetNodeList[CurrentNetNode->NodeID].Color == WHITE))
				{
					if (InitialDirection == UP)
						MainNetNodeList[CurrentNetNode->NodeID].Color = GRAY;
					else /* InitialDirection == DOWN */
						MainNetNodeList[CurrentNetNode->NodeID].Color = DARKGRAY;

					UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance = 
						CurrentDistance + 1;

					Error = UDAddConnectionToList(i, CurrentNetNode->NodeID, FirstConnectionID, CurrentDistance+1);
					if (Error != NOERROR)
					{
						printf("ERROR: Adding Connection to List in 'GenerateUpDownRoutingTable'\n");
						return(ERROR);
					}
		
					Error = AddToUDQueue(CurrentNetNode->NodeID, NewDirection, FirstConnectionID, InitialDirection, CurrentDistance+1);
					if (Error != NOERROR)
					{
						printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
						return(ERROR);
					}
				}
				else if ((LinkEligible) && (MainNetNodeList[CurrentNetNode->NodeID].Color == GRAY))
				{
					if (InitialDirection == DOWN)
					{
						MainNetNodeList[CurrentNetNode->NodeID].Color = DARKGRAY;
						UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance = CurrentDistance + 1; 

						/* alternative path with same distance */
						Error = UDAddConnectionToList(i, CurrentNetNode->NodeID, FirstConnectionID, CurrentDistance+1);
						if (Error != NOERROR)
						{
							printf("ERROR: Adding Connection to List in 'GenerateUpDownRoutingTable'\n");
							return(ERROR);
						}

						Error = AddToUDQueue(CurrentNetNode->NodeID, NewDirection, FirstConnectionID, InitialDirection, CurrentDistance+1);
						if (Error != NOERROR)
						{
							printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
							return(ERROR);
						}
					}
					else /* InitialDirection == UP */
					{
						if (UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance <  
								(CurrentDistance + 1))
						{
							/* ignore since a shorter path has already been found */
							if (NewDirection == UP)
							{
								Error = AddToUDQueue(CurrentNetNode->NodeID, NewDirection, FirstConnectionID, InitialDirection, CurrentDistance+1);
								if (Error != NOERROR)
								{
									printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
									return(ERROR);
								}
							}
						}
						else if (UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance ==  
								(CurrentDistance + 1))
						{
							/* alternative path with same distance */
							Error = UDAddConnectionToList(i, CurrentNetNode->NodeID, FirstConnectionID, CurrentDistance+1);
							if (Error != NOERROR)
							{
								printf("ERROR: Adding Connection to List in 'GenerateUpDownRoutingTable'\n");
								return(ERROR);
							}

							Error = AddToUDQueue(CurrentNetNode->NodeID, NewDirection, FirstConnectionID, InitialDirection, CurrentDistance+1);
							if (Error != NOERROR)
							{
								printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
								return(ERROR);
							}
						}
						else
							/* if (UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance >  
							   (UDLookUp[(i * IRNUMOFNODES) + CurrentNodeID].MinDistance + 1))
							 */
						{
							/* this case should never happen in BFS --- so report and return error */
							printf("ERROR: Found a shorter path unexpectedly in 'GenerateUpDownRoutingTable'\n");
							return(ERROR);

						}
					}
				}
				else if ((LinkEligible) && (MainNetNodeList[CurrentNetNode->NodeID].Color == DARKGRAY))
				{
					if (InitialDirection == DOWN)
					{
						if (UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance <  
								(CurrentDistance + 1))
						{
							/* ignore since a shorter path has already been found */
							MainNetNodeList[CurrentNodeID].Color = BLACK;
						}
						else if (UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance ==  
								(CurrentDistance + 1))
						{
							/* alternative path with same distance */
							Error = UDAddConnectionToList(i, CurrentNetNode->NodeID, FirstConnectionID, CurrentDistance+1);
							if (Error != NOERROR)
							{
								printf("ERROR: Adding Connection to List in 'GenerateUpDownRoutingTable'\n");
								return(ERROR);
							}

							Error = AddToUDQueue(CurrentNetNode->NodeID, NewDirection, FirstConnectionID, InitialDirection, CurrentDistance+1);
							if (Error != NOERROR)
							{
								printf("ERROR: Adding to UDQueue in 'GenerateUpDownRoutingTable'\n");
								return(ERROR);
							}
						}
						else
							/* if (UDLookUp[(i * IRNUMOFNODES) + CurrentNetNode->NodeID].MinDistance >  
							   (UDLookUp[(i * IRNUMOFNODES) + CurrentNodeID].MinDistance + 1))
							 */
						{
							/* this case should never happen in BFS --- so report and return error */
							printf("ERROR: Found a shorter path unexpectedly in 'GenerateUpDownRoutingTable'\n");
							return(ERROR);

						}
					}
				}

				CurrentNetNode = CurrentNetNode->NextNetNode;
			}
		}
	}

	free(UDQueue);
	return(NOERROR);
}	


/*
*	UpDownTableLookUp
*
*	Returns the list of connection (a circular linked 
*	list) of the Up/Down table entry specified via 
*	(CurrentNode, DestNode). This is intended
*	to be used for Up/Down-based routing algorithms
*	(i.e., Up/Down, Up/Down w/ Duato's)
*
*/

TABLEENTRYPTR	UpDownTableLookUp(CurrentNode, DestNode)
int	CurrentNode;
int	DestNode;
{
	return((TABLEENTRYPTR) &UDLookUp[(CurrentNode * IRNUMOFNODES) + DestNode]);
}	

/*** Hong ***/
double drand48();

int GetDestNodeFromConnectionID(int nd, int connectionID) {
  NETNODEPTR CurrentNetNode;

  CurrentNetNode = MainNetNodeList[nd].FirstNetNode;
  while (CurrentNetNode != NULL_) {
    if (CurrentNetNode->Connection == connectionID) return CurrentNetNode->NodeID;
    CurrentNetNode = CurrentNetNode->NextNetNode;
  }
  return -1;
}

int GetNodeAtDistance(int currentNode, int distance) {
  int numNodes, i;

  numNodes = 0;
  for (i=0; i<IRNUMOFNODES; i++) {
    if (i == currentNode) continue;
    if (LookUp[currentNode * IRNUMOFNODES + i].MinDistance <= distance)
      numNodes++;
  }
  if (numNodes == 0)
    return -1;

  numNodes = (int) drand48() * numNodes;
  for (i=0; i<IRNUMOFNODES; i++) {
    if (i == currentNode) continue;
    if (LookUp[currentNode * IRNUMOFNODES + i].MinDistance <= distance) {
      if (numNodes == 0) {
        return i;
      }
      numNodes--;
    }
  }
  return -1;
}
