/* 
*	Module: Irregular Network Support Header: 'irnet.h'
*/

/*
*	Constants (General Purpose)
*/
 
/*
*	Constants (Used for Breadth First Search)
*/

#define		WHITE		0
#define		GRAY		1
#define		DARKGRAY	2
#define		BLACK		3

/*
*	Constants (Used to define different types of routing tables)
*/

#define		STATIC		1
#define		ADAPTIVE	2
#define		UP_DOWN		3


/*
*	Constants (Used for Up/Down Routing Support)
*/

#define		UNDEFINED_LEVEL		-9

#define		UNDEFINED_DIRECTION	0
#define		UP			2
#define		DOWN			3

/* #define		UNDEFINED_CONNECTION	-1 */
#define		MAXUDQUEUE		5000

#define		UNDEFINED_CHANNEL	UNDEFINED_DIRECTION
#define		ADAPTIVE_CHANNEL	1
#define		UP_CHANNEL		UP
#define		DOWN_CHANNEL		DOWN


/*
*	Constants 
*
*/

/*
*----------------------------------------------------------------------------------
*/

#define		MAXNODES	256 		/*  Maximum number of network nodes allowed  */


/*
*
*	Type definitions for Deadlock Detection
*
*/

/*
*	These structures are used for maintaining adjancency list representation of irregular network,
*	and are used for analyzing topology and creating routing tables. 
*/

typedef struct NETNODE 
{
	int	NodeID;			/*  Identifies the Network Node */
	int	PhysLink;		/*  Indicates the Physical Link which connects this node */
	int	Connection;		/*  Connection ID for the Physical Link */
	struct  NETNODE  *NextNetNode;	/*  Points to the next VC being used (if any) */

} NETNODE, *NETNODEPTR;


typedef struct MAINNETNODE 
{
	int	NodeID;			/*  Identifies the Network Node */
	int	Degree;			/*  Indicates Network Degree (number of adjacent nodes)  */
	int	Color;			/*  Color --- used for Breadth First Search during routing tabe generation  */
	int	Level;			/*  Level --- used to specify level of node within Min. Span. Tree (for Up/Down) */
	int	HotSpot;		/*  Hotspot --- used to indicate if this is a server node in Sot Spot traffic */
	struct  NETNODE  *FirstNetNode;	/*  Points to the first adjacent node */
} MAINNETNODE, *MAINNETNODEPTR;


typedef struct TABLEENTRY 
{
	int	MinDistance;	           	/*  Indicates Corrsponding Path (Minimum) Distance between Current and Destination Nodes  */
						/*  May not always be actual Min. Distance for routing algorithms that allow non-min. paths */
	int	NumOfConnections;	        /*  Indicates Number of Connections available to get to next hop (for adaptive routing) */
	int	StaticConnectionID;	        /*  Identifies the connection to be used (for static routing only) */
	struct  CONNLISTITEM  *FirstConnection;	/*  Points to the first in a linked list of connections */
} TABLEENTRY, *TABLEENTRYPTR;

typedef struct CONNLISTITEM 
{
	int	ConnectionID;			/*  Identifies the connection */
	int	Distance;	       		/*  Indicates distance of the path along this connection (to be used to    */
						/*    support paths of varying lengths for a given source/dest. pair)      */
	struct  CONNLISTITEM  *NextConnection;	/*  Points to the first in a linked list */
} CONNLISTITEM, *CONNLISTITEMPTR;


typedef struct UDQUEUEENTRY 
{
	int	NodeID;				/*  Identifies the Node */
	int	Direction;	       		/*  Indicates direction (Up/Down) taken to reach node  */
	int	ConnectionID;			/*  Connection ID to be used */
	int	InitialDirection;		/*  Initial direction of the path */
	int	Distance;			/*  Path distance up to current node */
} UDQUEUEENTRY, *UDQUEUEENTRYPTR;


/* 
*
*	  I N I T I A L I Z A T I O N   R O U T I N E S 
*
*/

void	InitMainNetNodeList();

void	FreeMainNetNodeListEntries();

void	InitColor();


/* 
*
*	  P R I N T   ( D E B U G G I N G )   R O U T I N E S 
*
*/

void	PrintIRNetAdjacencyList();

void	PrintStaticRoutingTable();


/* 
*
*	  S T A T I S T I C S   P R I N T I N G   R O U T I N E S 
*
*/

void	PrintIRNetStats();


/* 
*
*	  F I L E   I N P U T / O U T P U T   R O U T I N E S 
*
*/

int WriteNetworktoFile(char *OutputFileName);

int WriteRoutingTabletoFile(char *OutputFileName, int TableType);

int WriteNetworktoDaVinciFile(char *DaVinciFileName);


/* 
*
*	  N E T W O R K   G E N E R A T I O N   R O U T I N E S 
*
*/

int GenerateNetwork(int NumOfNodes, int NumOfLinks, int MinDegree, int MaxDegree, int AllowDuplicates);

int GenerateNetworkFromFile(char *InputFileName, int *NumOfNodes, int *NumOfLinks);


/* 
*
*	  A D J A C E N C Y   L I S T   M A N A G E M E N T   R O U T I N E S 
*
*/

int	AddNetNodeToList(int NodeID, int NewNodeID);

int	FindEntryInAdjacencyList(int ListIndex, int EntryNodeID);

void 	SetPhysicalLink(int ListIndex, int EntryNodeID, int PhysicalLinkID);

void	SetUpLinks(int NumOfNodes);

void	SetConnections(int NumOfNodes);


/* 
*
*	  F I F O   Q U E U E   M A N A G E M E N T   R O U T I N E S 
*
*/

int	AllocateQueue();

void	InitializeQueue();

int	AddToQueue(int NewQueueEntry);

int	RemoveFromQueue();

int	QueueEmpty();


/* 
*
*	  R O U T I N G   T A B L E   M A N A G E M E N T   R O U T I N E S 
*
*/

int	AllocateRoutingTable();

void	FreeRoutingTable();

void	FreeRoutingTableList(int i, int j);

void	InitializeRoutingTable();

int	AddConnectionToList(int CurrentNode, int DestNode, int NewConnID, int NewDistance);

int	AddConnectionListToList(int FromCurrentNode, int FromDestNode, int ToCurrentNode, int ToDestNode);

int	GenerateStaticRoutingTableFromFile(char *InputFileName, int *NumOfEntries);

void	DetermineStaticRoutingConnections();

int	StaticTableLookUp(int CurrentNode, int DestNode);

int	GenerateMinTFARoutingTable();

TABLEENTRYPTR	AdaptiveTableLookUp(int CurrentNode, int DestNode);


/* 
*
*	  D A T A   S E R V I C E S   R O U T I N E S 
*
*/

int GetNodeDegree(int NodeID);

int GetDistance(int SourceNodeID, int DestNodeID);

int GetNextHotSpot(int UpperBoundNodeID);


/* 
*
*	  U P / D O W N   R O U T I N E S 
*
*/

int	DetermineRootOfMST();

int	DefineMSTLevels(int RootNodeID);


/* 
*
*	  U P / D O W N   F I F O   Q U E U E   M A N A G E M E N T   R O U T I N E S 
*
*/

int	AllocateUDQueue();

void	InitUDQueue();

int	AddToUDQueue(int NodeID, int Direction, int ConnectionID, int InitialDirection, int Distance);

void	RemoveFromUDQueue(int *NodeID, int *Direction, int *ConnectionID, int *InitialDirection, int *Distance);

int	UDQueueEmpty();


/* 
*
*	   U P / D O W N   R O U T I N G   T A B L E   M A N A G E M E N T   R O U T I N E S 
*
*/

int	AllocateUDRoutingTable();

void	FreeUDRoutingTable();

void	FreeUDRoutingTableList(int i, int j);

void	InitializeUDRoutingTable();

int	UDAddConnectionToList(int CurrentNode, int DestNode, int NewConnID, int NewDistance);

int	UDAddConnectionListToList(int FromCurrentNode, int FromDestNode, int ToCurrentNode, int ToDestNode);

int	DetermineDirection(int CurrentNodeID, int NextNodeID);

int	DetermineLinkEligibility(int PreviousDirection, int NewDirection);

int	GenerateUpDownRoutingTable();

TABLEENTRYPTR	UpDownTableLookUp(int CurrentNode, int DestNode);

int	GetNodeFromConnectionID(int node, int connectionID);
